Namespace NeuralTimeSeries
	Friend Class NeuralTimeSeriesProgram
		Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin neural network times series demo")
            Console.WriteLine("Goal is to predict airline passengers over time ")
            Console.WriteLine("Data from January 1949 to December 1960 " & vbCrLf)

            Dim trainData()() As Double = GetAirlineData()
			trainData = Normalize(trainData)
			Console.WriteLine("Normalized training data:")
			ShowMatrix(trainData, 5, 2, True) ' first 5 rows, 2 decimals, show indices

            Dim numInput = 4 ' number predictors
            Dim numHidden = 12
            Dim numOutput = 1 ' regression

			Console.WriteLine("Creating a " & numInput & "-" & numHidden & "-" & numOutput & " neural network")
			Dim nn As New NeuralNetwork(numInput, numHidden, numOutput)

            Dim maxEpochs = 10000
            Dim learnRate As Double = 0.01
            Console.WriteLine(vbCrLf & "Setting maxEpochs = " & maxEpochs)
            Console.WriteLine("Setting learnRate = " & learnRate.ToString("F2"))

            Console.WriteLine(vbCrLf & "Starting training")
            Dim weights() As Double = nn.Train(trainData, maxEpochs, learnRate)
			Console.WriteLine("Done")
            Console.WriteLine(vbCrLf & "Final neural network model weights and biases:" & vbCrLf)
            ShowVector(weights, 2, 10, True)

			Dim trainAcc As Double = nn.Accuracy(trainData, 0.30) ' within 30
            Console.WriteLine(vbCrLf & "Model accuracy (+/- 30) on training data = " & trainAcc.ToString("F4"))

            Dim predictors() As Double = { 5.08, 4.61, 3.90, 4.32 }
			Dim forecast() As Double = nn.ComputeOutputs(predictors) ' 4.33362252510741
            Console.WriteLine(vbCrLf & "Predicted passengers for January 1961 (t=145): ")
            Console.WriteLine((forecast(0) * 100).ToString("F0"))

            Console.WriteLine(vbCrLf & "End time series demo" & vbCrLf)
            Console.ReadLine()
		End Sub ' Main

        Private Shared Function Normalize(ByVal data()() As Double) As Double()()
            ' divide all by 100.0
            Dim rows = data.Length
            Dim cols = data(0).Length
            Dim result(rows - 1)() As Double
            For i = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i

            For i = 0 To rows - 1
                For j = 0 To cols - 1
                    result(i)(j) = data(i)(j) / 100.0
                Next j
            Next i
            Return result
        End Function

        Private Shared Function GetAirlineData() As Double()()
			Dim airData(139)() As Double
			airData(0) = New Double() { 112, 118, 132, 129, 121 }
			airData(1) = New Double() { 118, 132, 129, 121, 135 }
			airData(2) = New Double() { 132, 129, 121, 135, 148 }
			airData(3) = New Double() { 129, 121, 135, 148, 148 }
			airData(4) = New Double() { 121, 135, 148, 148, 136 }
			airData(5) = New Double() { 135, 148, 148, 136, 119 }
			airData(6) = New Double() { 148, 148, 136, 119, 104 }
			airData(7) = New Double() { 148, 136, 119, 104, 118 }
			airData(8) = New Double() { 136, 119, 104, 118, 115 }
			airData(9) = New Double() { 119, 104, 118, 115, 126 }
			airData(10) = New Double() { 104, 118, 115, 126, 141 }
			airData(11) = New Double() { 118, 115, 126, 141, 135 }
			airData(12) = New Double() { 115, 126, 141, 135, 125 }
			airData(13) = New Double() { 126, 141, 135, 125, 149 }
			airData(14) = New Double() { 141, 135, 125, 149, 170 }
			airData(15) = New Double() { 135, 125, 149, 170, 170 }
			airData(16) = New Double() { 125, 149, 170, 170, 158 }
			airData(17) = New Double() { 149, 170, 170, 158, 133 }
			airData(18) = New Double() { 170, 170, 158, 133, 114 }
			airData(19) = New Double() { 170, 158, 133, 114, 140 }
			airData(20) = New Double() { 158, 133, 114, 140, 145 }
			airData(21) = New Double() { 133, 114, 140, 145, 150 }
			airData(22) = New Double() { 114, 140, 145, 150, 178 }
			airData(23) = New Double() { 140, 145, 150, 178, 163 }
			airData(24) = New Double() { 145, 150, 178, 163, 172 }
			airData(25) = New Double() { 150, 178, 163, 172, 178 }
			airData(26) = New Double() { 178, 163, 172, 178, 199 }
			airData(27) = New Double() { 163, 172, 178, 199, 199 }
			airData(28) = New Double() { 172, 178, 199, 199, 184 }
			airData(29) = New Double() { 178, 199, 199, 184, 162 }
			airData(30) = New Double() { 199, 199, 184, 162, 146 }
			airData(31) = New Double() { 199, 184, 162, 146, 166 }
			airData(32) = New Double() { 184, 162, 146, 166, 171 }
			airData(33) = New Double() { 162, 146, 166, 171, 180 }
			airData(34) = New Double() { 146, 166, 171, 180, 193 }
			airData(35) = New Double() { 166, 171, 180, 193, 181 }
			airData(36) = New Double() { 171, 180, 193, 181, 183 }
			airData(37) = New Double() { 180, 193, 181, 183, 218 }
			airData(38) = New Double() { 193, 181, 183, 218, 230 }
			airData(39) = New Double() { 181, 183, 218, 230, 242 }
			airData(40) = New Double() { 183, 218, 230, 242, 209 }
			airData(41) = New Double() { 218, 230, 242, 209, 191 }
			airData(42) = New Double() { 230, 242, 209, 191, 172 }
			airData(43) = New Double() { 242, 209, 191, 172, 194 }
			airData(44) = New Double() { 209, 191, 172, 194, 196 }
			airData(45) = New Double() { 191, 172, 194, 196, 196 }
			airData(46) = New Double() { 172, 194, 196, 196, 236 }
			airData(47) = New Double() { 194, 196, 196, 236, 235 }
			airData(48) = New Double() { 196, 196, 236, 235, 229 }
			airData(49) = New Double() { 196, 236, 235, 229, 243 }
			airData(50) = New Double() { 236, 235, 229, 243, 264 }
			airData(51) = New Double() { 235, 229, 243, 264, 272 }
			airData(52) = New Double() { 229, 243, 264, 272, 237 }
			airData(53) = New Double() { 243, 264, 272, 237, 211 }
			airData(54) = New Double() { 264, 272, 237, 211, 180 }
			airData(55) = New Double() { 272, 237, 211, 180, 201 }
			airData(56) = New Double() { 237, 211, 180, 201, 204 }
			airData(57) = New Double() { 211, 180, 201, 204, 188 }
			airData(58) = New Double() { 180, 201, 204, 188, 235 }
			airData(59) = New Double() { 201, 204, 188, 235, 227 }
			airData(60) = New Double() { 204, 188, 235, 227, 234 }
			airData(61) = New Double() { 188, 235, 227, 234, 264 }
			airData(62) = New Double() { 235, 227, 234, 264, 302 }
			airData(63) = New Double() { 227, 234, 264, 302, 293 }
			airData(64) = New Double() { 234, 264, 302, 293, 259 }
			airData(65) = New Double() { 264, 302, 293, 259, 229 }
			airData(66) = New Double() { 302, 293, 259, 229, 203 }
			airData(67) = New Double() { 293, 259, 229, 203, 229 }
			airData(68) = New Double() { 259, 229, 203, 229, 242 }
			airData(69) = New Double() { 229, 203, 229, 242, 233 }
			airData(70) = New Double() { 203, 229, 242, 233, 267 }
			airData(71) = New Double() { 229, 242, 233, 267, 269 }
			airData(72) = New Double() { 242, 233, 267, 269, 270 }
			airData(73) = New Double() { 233, 267, 269, 270, 315 }
			airData(74) = New Double() { 267, 269, 270, 315, 364 }
			airData(75) = New Double() { 269, 270, 315, 364, 347 }
			airData(76) = New Double() { 270, 315, 364, 347, 312 }
			airData(77) = New Double() { 315, 364, 347, 312, 274 }
			airData(78) = New Double() { 364, 347, 312, 274, 237 }
			airData(79) = New Double() { 347, 312, 274, 237, 278 }
			airData(80) = New Double() { 312, 274, 237, 278, 284 }
			airData(81) = New Double() { 274, 237, 278, 284, 277 }
			airData(82) = New Double() { 237, 278, 284, 277, 317 }
			airData(83) = New Double() { 278, 284, 277, 317, 313 }
			airData(84) = New Double() { 284, 277, 317, 313, 318 }
			airData(85) = New Double() { 277, 317, 313, 318, 374 }
			airData(86) = New Double() { 317, 313, 318, 374, 413 }
			airData(87) = New Double() { 313, 318, 374, 413, 405 }
			airData(88) = New Double() { 318, 374, 413, 405, 355 }
			airData(89) = New Double() { 374, 413, 405, 355, 306 }
			airData(90) = New Double() { 413, 405, 355, 306, 271 }
			airData(91) = New Double() { 405, 355, 306, 271, 306 }
			airData(92) = New Double() { 355, 306, 271, 306, 315 }
			airData(93) = New Double() { 306, 271, 306, 315, 301 }
			airData(94) = New Double() { 271, 306, 315, 301, 356 }
			airData(95) = New Double() { 306, 315, 301, 356, 348 }
			airData(96) = New Double() { 315, 301, 356, 348, 355 }
			airData(97) = New Double() { 301, 356, 348, 355, 422 }
			airData(98) = New Double() { 356, 348, 355, 422, 465 }
			airData(99) = New Double() { 348, 355, 422, 465, 467 }
			airData(100) = New Double() { 355, 422, 465, 467, 404 }
			airData(101) = New Double() { 422, 465, 467, 404, 347 }
			airData(102) = New Double() { 465, 467, 404, 347, 305 }
			airData(103) = New Double() { 467, 404, 347, 305, 336 }
			airData(104) = New Double() { 404, 347, 305, 336, 340 }
			airData(105) = New Double() { 347, 305, 336, 340, 318 }
			airData(106) = New Double() { 305, 336, 340, 318, 362 }
			airData(107) = New Double() { 336, 340, 318, 362, 348 }
			airData(108) = New Double() { 340, 318, 362, 348, 363 }
			airData(109) = New Double() { 318, 362, 348, 363, 435 }
			airData(110) = New Double() { 362, 348, 363, 435, 491 }
			airData(111) = New Double() { 348, 363, 435, 491, 505 }
			airData(112) = New Double() { 363, 435, 491, 505, 404 }
			airData(113) = New Double() { 435, 491, 505, 404, 359 }
			airData(114) = New Double() { 491, 505, 404, 359, 310 }
			airData(115) = New Double() { 505, 404, 359, 310, 337 }
			airData(116) = New Double() { 404, 359, 310, 337, 360 }
			airData(117) = New Double() { 359, 310, 337, 360, 342 }
			airData(118) = New Double() { 310, 337, 360, 342, 406 }
			airData(119) = New Double() { 337, 360, 342, 406, 396 }
			airData(120) = New Double() { 360, 342, 406, 396, 420 }
			airData(121) = New Double() { 342, 406, 396, 420, 472 }
			airData(122) = New Double() { 406, 396, 420, 472, 548 }
			airData(123) = New Double() { 396, 420, 472, 548, 559 }
			airData(124) = New Double() { 420, 472, 548, 559, 463 }
			airData(125) = New Double() { 472, 548, 559, 463, 407 }
			airData(126) = New Double() { 548, 559, 463, 407, 362 }
			airData(127) = New Double() { 559, 463, 407, 362, 405 }
			airData(128) = New Double() { 463, 407, 362, 405, 417 }
			airData(129) = New Double() { 407, 362, 405, 417, 391 }
			airData(130) = New Double() { 362, 405, 417, 391, 419 }
			airData(131) = New Double() { 405, 417, 391, 419, 461 }
			airData(132) = New Double() { 417, 391, 419, 461, 472 }
			airData(133) = New Double() { 391, 419, 461, 472, 535 }
			airData(134) = New Double() { 419, 461, 472, 535, 622 }
			airData(135) = New Double() { 461, 472, 535, 622, 606 }
			airData(136) = New Double() { 472, 535, 622, 606, 508 }
			airData(137) = New Double() { 535, 622, 606, 508, 461 }
			airData(138) = New Double() { 622, 606, 508, 461, 390 }
			airData(139) = New Double() { 606, 508, 461, 390, 432 }
			Return airData
		End Function

		Private Shared Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer, ByVal decimals As Integer, ByVal indices As Boolean)
            Dim len = matrix.Length.ToString().Length
            For i = 0 To numRows - 1
                If indices = True Then
                    Console.Write("[" & i.ToString().PadLeft(len) & "]  ")
                End If
                For j = 0 To matrix(i).Length - 1
                    Dim v As Double = matrix(i)(j)
                    If v >= 0.0 Then
                        Console.Write(" ") ' '+'
                    End If
                    Console.Write(v.ToString("F" & decimals) & "  ")
                Next j
                Console.WriteLine("")
            Next i

            If numRows < matrix.Length Then
				Console.WriteLine(". . .")
                Dim lastRow = matrix.Length - 1
                If indices = True Then
					Console.Write("[" & lastRow.ToString().PadLeft(len) & "]  ")
				End If
                For j = 0 To matrix(lastRow).Length - 1
                    Dim v As Double = matrix(lastRow)(j)
                    If v >= 0.0 Then
                        Console.Write(" ") ' '+'
                    End If
                    Console.Write(v.ToString("F" & decimals) & "  ")
                Next j
            End If
            Console.WriteLine(vbCrLf)
        End Sub

		Private Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal lineLen As Integer, ByVal newLine As Boolean)
            For i = 0 To vector.Length - 1
                If i > 0 AndAlso i Mod lineLen = 0 Then
                    Console.WriteLine("")
                End If
                If vector(i) >= 0 Then
                    Console.Write(" ")
                End If
                Console.Write(vector(i).ToString("F" & decimals) & " ")
            Next i
            If newLine = True Then
				Console.WriteLine("")
			End If
		End Sub


	End Class ' Program

	Public Class NeuralNetwork
        Private numInput As Integer ' number input nodes
        Private numHidden As Integer
        Private numOutput As Integer

        Private iNodes() As Double
		Private ihWeights()() As Double ' input-hidden
		Private hBiases() As Double
		Private hNodes() As Double

		Private hoWeights()() As Double ' hidden-output
		Private oBiases() As Double
		Private oNodes() As Double

		Private rnd As Random

		Public Sub New(ByVal numInput As Integer, ByVal numHidden As Integer, ByVal numOutput As Integer)
			Me.numInput = numInput
			Me.numHidden = numHidden
			Me.numOutput = numOutput

			Me.iNodes = New Double(numInput - 1){}

			Me.ihWeights = MakeMatrix(numInput, numHidden, 0.0)
			Me.hBiases = New Double(numHidden - 1){}
			Me.hNodes = New Double(numHidden - 1){}

			Me.hoWeights = MakeMatrix(numHidden, numOutput, 0.0)
			Me.oBiases = New Double(numOutput - 1){}
			Me.oNodes = New Double(numOutput - 1){}

			Me.rnd = New Random(0)
			Me.InitializeWeights() ' all weights and biases
		End Sub ' ctor

		Private Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer, ByVal v As Double) As Double()() ' helper for ctor, Train
			Dim result(rows - 1)() As Double
            For r = 0 To result.Length - 1
                result(r) = New Double(cols - 1) {}
            Next r
            For i = 0 To rows - 1
                For j = 0 To cols - 1
                    result(i)(j) = v
                Next j
            Next i
            Return result
		End Function

		Private Sub InitializeWeights() ' helper for ctor
            ' initialize weights and biases to small random values
            Dim numWeights = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
            Dim initialWeights(numWeights - 1) As Double
            For i = 0 To initialWeights.Length - 1
                initialWeights(i) = (0.001 - 0.0001) * rnd.NextDouble() + 0.0001
            Next i
            Me.SetWeights(initialWeights)
		End Sub

		Public Sub SetWeights(ByVal weights() As Double)
            ' copy serialized weights and biases in weights[] array
            ' to i-h weights, i-h biases, h-o weights, h-o biases
            Dim numWeights = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
            If weights.Length <> numWeights Then
				Throw New Exception("Bad weights array in SetWeights")
			End If

            Dim k = 0 ' points into weights param

            For i = 0 To numInput - 1
                For j = 0 To numHidden - 1
                    ihWeights(i)(j) = weights(k)
                    k += 1
                Next j
            Next i
            For i = 0 To numHidden - 1
                hBiases(i) = weights(k)
                k += 1
            Next i
            For i = 0 To numHidden - 1
                For j = 0 To numOutput - 1
                    hoWeights(i)(j) = weights(k)
                    k += 1
                Next j
            Next i
            For i = 0 To numOutput - 1
                oBiases(i) = weights(k)
                k += 1
            Next i
        End Sub

		Public Function GetWeights() As Double()
            Dim numWeights = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
            Dim result(numWeights - 1) As Double
            Dim k = 0
            For i = 0 To ihWeights.Length - 1
                For j = 0 To ihWeights(0).Length - 1
                    result(k) = ihWeights(i)(j)
                    k += 1
                Next j
            Next i
            For i = 0 To hBiases.Length - 1
                result(k) = hBiases(i)
                k += 1
            Next i
            For i = 0 To hoWeights.Length - 1
                For j = 0 To hoWeights(0).Length - 1
                    result(k) = hoWeights(i)(j)
                    k += 1
                Next j
            Next i
            For i = 0 To oBiases.Length - 1
                result(k) = oBiases(i)
                k += 1
            Next i
            Return result
		End Function

		Public Function ComputeOutputs(ByVal xValues() As Double) As Double()
			Dim hSums(numHidden - 1) As Double ' hidden nodes sums scratch array
			Dim oSums(numOutput - 1) As Double ' output nodes sums

            For i = 0 To xValues.Length - 1 ' copy x-values to inputs
                Me.iNodes(i) = xValues(i)
            Next i

            For j = 0 To numHidden - 1 ' compute i-h sum of weights * inputs
                For i = 0 To numInput - 1
                    hSums(j) += Me.iNodes(i) * Me.ihWeights(i)(j) ' note +=
                Next i
            Next j

            For i = 0 To numHidden - 1 ' add biases to hidden sums
                hSums(i) += Me.hBiases(i)
            Next i

            For i = 0 To numHidden - 1 ' apply activation
                Me.hNodes(i) = HyperTan(hSums(i)) ' hard-coded
            Next i

            For j = 0 To numOutput - 1 ' compute h-o sum of weights * hOutputs
                For i = 0 To numHidden - 1
                    oSums(j) += hNodes(i) * hoWeights(i)(j)
                Next i
            Next j

            For i = 0 To numOutput - 1 ' add biases to output sums
                oSums(i) += oBiases(i)
            Next i

            Array.Copy(oSums, Me.oNodes, oSums.Length) ' really only 1 value

			Dim retResult(numOutput - 1) As Double ' could define a GetOutputs
			Array.Copy(Me.oNodes, retResult, retResult.Length)
			Return retResult
		End Function

		Private Shared Function HyperTan(ByVal x As Double) As Double
			If x < -20.0 Then ' approximation is correct to 30 decimals
				Return -1.0
			ElseIf x > 20.0 Then
				Return 1.0
			Else
				Return Math.Tanh(x)
			End If
		End Function

		Private Shared Function LogSig(ByVal x As Double) As Double
			If x < -20.0 Then ' approximation
				Return 0.0
			ElseIf x > 20.0 Then
				Return 1.0
			Else
				Return 1.0 / (1.0 + Math.Exp(x))
			End If
		End Function

		Public Function Train(ByVal trainData()() As Double, ByVal maxEpochs As Integer, ByVal learnRate As Double) As Double()
			' train using back-prop
			' back-prop specific arrays
			Dim hoGrads()() As Double = MakeMatrix(numHidden, numOutput, 0.0) ' hidden-to-output weight gradients
			Dim obGrads(numOutput - 1) As Double ' output bias gradients

			Dim ihGrads()() As Double = MakeMatrix(numInput, numHidden, 0.0) ' input-to-hidden weight gradients
			Dim hbGrads(numHidden - 1) As Double ' hidden bias gradients

			Dim oSignals(numOutput - 1) As Double ' local gradient output signals
			Dim hSignals(numHidden - 1) As Double ' local gradient hidden node signals

            Dim epoch = 0
            Dim xValues(numInput - 1) As Double ' inputs
			Dim tValues(numOutput - 1) As Double ' target values
			Dim derivative As Double = 0.0
			Dim errorSignal As Double = 0.0

			Dim sequence(trainData.Length - 1) As Integer
            For i = 0 To sequence.Length - 1
                sequence(i) = i
            Next i

            Dim errInterval = maxEpochs \ 5 ' interval to check error
            Do While epoch < maxEpochs
				epoch += 1

				If epoch Mod errInterval = 0 AndAlso epoch < maxEpochs Then
					Dim trainErr As Double = [Error](trainData)
					Console.WriteLine("epoch = " & epoch & "  error = " & trainErr.ToString("F4"))
				End If

				Shuffle(sequence) ' visit each training data in random order
                For ii = 0 To trainData.Length - 1
                    Dim idx = sequence(ii)
                    Array.Copy(trainData(idx), xValues, numInput)
                    Array.Copy(trainData(idx), numInput, tValues, 0, numOutput)
                    ComputeOutputs(xValues) ' copy xValues in, compute outputs

                    ' indices: i = inputs, j = hiddens, k = outputs

                    ' 1. compute output node signals (assumes softmax)
                    For k = 0 To numOutput - 1
                        errorSignal = tValues(k) - oNodes(k) ' Wikipedia uses (o-t)
                        derivative = 1.0 ' for Identity activation
                        oSignals(k) = errorSignal * derivative
                    Next k

                    ' 2. compute hidden-to-output weight gradients using output signals
                    For j = 0 To numHidden - 1
                        For k = 0 To numOutput - 1
                            hoGrads(j)(k) = oSignals(k) * hNodes(j)
                        Next k
                    Next j

                    ' 2b. compute output bias gradients using output signals
                    For k = 0 To numOutput - 1
                        obGrads(k) = oSignals(k) * 1.0 ' dummy assoc. input value
                    Next k

                    ' 3. compute hidden node signals
                    For j = 0 To numHidden - 1
                        derivative = (1 + hNodes(j)) * (1 - hNodes(j)) ' for tanh
                        Dim sum As Double = 0.0 ' need sums of output signals times hidden-to-output weights
                        For k = 0 To numOutput - 1
                            sum += oSignals(k) * hoWeights(j)(k) ' represents error signal
                        Next k
                        hSignals(j) = derivative * sum
                    Next j

                    ' 4. compute input-hidden weight gradients
                    For i = 0 To numInput - 1
                        For j = 0 To numHidden - 1
                            ihGrads(i)(j) = hSignals(j) * iNodes(i)
                        Next j
                    Next i

                    ' 4b. compute hidden node bias gradients
                    For j = 0 To numHidden - 1
                        hbGrads(j) = hSignals(j) * 1.0 ' dummy 1.0 input
                    Next j

                    ' == update weights and biases

                    ' update input-to-hidden weights
                    For i = 0 To numInput - 1
                        For j = 0 To numHidden - 1
                            Dim delta As Double = ihGrads(i)(j) * learnRate
                            ihWeights(i)(j) += delta ' would be -= if (o-t)
                        Next j
                    Next i

                    ' update hidden biases
                    For j = 0 To numHidden - 1
                        Dim delta As Double = hbGrads(j) * learnRate
                        hBiases(j) += delta
                    Next j

                    ' update hidden-to-output weights
                    For j = 0 To numHidden - 1
                        For k = 0 To numOutput - 1
                            Dim delta As Double = hoGrads(j)(k) * learnRate
                            hoWeights(j)(k) += delta
                        Next k
                    Next j

                    ' update output node biases
                    For k = 0 To numOutput - 1
                        Dim delta As Double = obGrads(k) * learnRate
                        oBiases(k) += delta
                    Next k

                Next ii ' each training item

            Loop ' while
			Dim bestWts() As Double = GetWeights()
			Return bestWts
		End Function ' Train

		Private Sub Shuffle(ByVal sequence() As Integer) ' instance method
            For i = 0 To sequence.Length - 1
                Dim r = Me.rnd.Next(i, sequence.Length)
                Dim tmp = sequence(r)
                sequence(r) = sequence(i)
                sequence(i) = tmp
            Next i
        End Sub ' Shuffle

		Private Function [Error](ByVal trainData()() As Double) As Double
			' average squared error per training item
			Dim sumSquaredError As Double = 0.0
			Dim xValues(numInput - 1) As Double ' first numInput values in trainData
			Dim tValues(numOutput - 1) As Double ' last numOutput values

            ' walk thru each training case. looks like (6.9 3.2 5.7 2.3) (0 0 1)
            For i = 0 To trainData.Length - 1
                Array.Copy(trainData(i), xValues, numInput)
                Array.Copy(trainData(i), numInput, tValues, 0, numOutput) ' get target values
                Dim yValues() As Double = Me.ComputeOutputs(xValues) ' outputs using current weights
                For j = 0 To numOutput - 1
                    Dim err As Double = tValues(j) - yValues(j)
                    sumSquaredError += err * err
                Next j
            Next i
            Return sumSquaredError / trainData.Length
		End Function ' MeanSquaredError

		Public Function Accuracy(ByVal testData()() As Double, ByVal howClose As Double) As Double
            ' percentage correct using winner-takes all
            Dim numCorrect = 0
            Dim numWrong = 0
            Dim xValues(numInput - 1) As Double ' inputs
			Dim tValues(numOutput - 1) As Double ' targets
			Dim yValues() As Double ' computed Y

            For i = 0 To testData.Length - 1
                Array.Copy(testData(i), xValues, numInput) ' get x-values
                Array.Copy(testData(i), numInput, tValues, 0, numOutput) ' get t-values
                yValues = Me.ComputeOutputs(xValues)

                If Math.Abs(yValues(0) - tValues(0)) < howClose Then ' within 30
                    numCorrect += 1
                Else
                    numWrong += 1
                End If

            Next i
            Return (numCorrect * 1.0) / (numCorrect + numWrong)
		End Function

	End Class ' class NeuralNetwork

End Namespace ' ns
